from typing import Literal, TypedDict

from openai.types.chat import ChatCompletionUserMessageParam
from pydantic import BaseModel, Field

from tfy_assistant_framework.assistant import (
    AssistantUpdateToUser,
    send_update,
)
from tfy_assistant_framework.swarm import (
    Agent,
    AgentResponseValidationError,
    ContextVarType,
    get_response_from_structured_agent,
)


# Tools and state setup
class AccountInfo(TypedDict):
    balance: float
    name: str


ACCOUNTS: dict[int, AccountInfo] = {
    1: {"balance": 1000.0, "name": "John"},
    2: {"balance": 500.0, "name": "Alice"},
}


def is_account_valid(account_id: int) -> bool:
    return account_id in ACCOUNTS


def get_balance(context_variables: ContextVarType) -> float:
    return ACCOUNTS[context_variables["initiator_account_id"]]["balance"]


def make_transfer(
    context_variables: ContextVarType, to_account: int, amount: float
) -> str:
    sender = ACCOUNTS[context_variables["initiator_account_id"]]
    recipient = ACCOUNTS.get(to_account)

    if not recipient:
        raise ValueError("Recipient not found")
    if amount > sender["balance"]:
        raise ValueError("Insufficient funds")

    sender["balance"] -= amount
    recipient["balance"] += amount
    return f"Transferred ${amount} to {recipient['name']}"


# Agents Setup
class TransferIntent(BaseModel):
    intent_type: Literal["transfer"] = Field(
        ..., description="Intent type, only accepts 'transfer'"
    )
    amount: float = Field(..., description="Transfer amount", gt=0)
    to_account: int = Field(..., description="Recipient account ID")


class ExitIntent(BaseModel):
    intent_type: Literal["exit"] = Field(
        ..., description="Intent type, only accepts 'exit'"
    )


class IntentAgentResponse(BaseModel):
    intent: TransferIntent | ExitIntent


class TransferResultResponse(BaseModel):
    message: str = Field(..., description="Description of the transfer")


class IntentAgent(Agent[IntentAgentResponse]):
    async def validate_response(
        self,
        response: IntentAgentResponse,
        context_variables: ContextVarType,
    ) -> AgentResponseValidationError | IntentAgentResponse:
        if response.intent.intent_type == "transfer":  # acts as a hard gaurd rail
            if not is_account_valid(response.intent.to_account):
                return AgentResponseValidationError(
                    error_message="Recipient account is not valid",
                )
            if (
                response.intent.amount
                > ACCOUNTS[context_variables["initiator_account_id"]]["balance"]
            ):
                return AgentResponseValidationError(
                    error_message="Insufficient funds",
                )

        return response


class TransferAgent(Agent[TransferResultResponse]): ...


intent_agent = IntentAgent(
    response_type=IntentAgentResponse,
    instructions="Interact with user and Identify banking intent from user request",
    functions=[get_balance, is_account_valid],
)

transfer_agent = TransferAgent(
    response_type=TransferResultResponse,
    instructions="Handle money transfers between accounts",
    functions=[make_transfer, get_balance],
)


class CustomerSupportRequest(BaseModel):
    account_id: int


# Run the customer support
async def run_customer_support_agent(body: CustomerSupportRequest) -> None:
    context_variables = {"initiator_account_id": body.account_id}

    response = await get_response_from_structured_agent(
        intent_agent, context_variables=context_variables, interactive=True
    )

    if response.intent.intent_type == "transfer":
        result = await get_response_from_structured_agent(
            transfer_agent,
            context_variables=context_variables,
            interactive=False,
            messages=[
                ChatCompletionUserMessageParam(
                    content=response.intent.model_dump_json(), role="user"
                )
            ],
        )
        await send_update(AssistantUpdateToUser(content=result.message))

    await send_update(AssistantUpdateToUser(content="Thank you for contacting us"))


async def main() -> None:
    await run_customer_support_agent(CustomerSupportRequest(account_id=1))


if __name__ == "__main__":
    import asyncio

    result = asyncio.run(main())
    print(result)



import uuid
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

import logfire
from fastapi import FastAPI, Response
from nats.js import JetStreamContext

from examples.customer_support.customer_support import (
    CustomerSupportRequest,
    run_customer_support_agent,
)
from tfy_assistant_framework.assistant import (
    AssistantServe,
    CreateAssistantTaskRunResult,
    NATSAssistantUpdateSink,
    PostAssistantTaskMessage,
)
from tfy_assistant_framework.nats_client import nats_connection

js: JetStreamContext
assistant_serve = AssistantServe()


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncIterator[None]:
    global js

    async with nats_connection() as (nc_, js):
        js = js

        async with assistant_serve.init(nc=nc_, app=app):
            yield


app = FastAPI(
    title="LLM Assistants Server",
    lifespan=lifespan,
    docs_url="/",
)


@app.post("/assistants/customer_support/task")
async def customer_support_task(
    body: CustomerSupportRequest,
) -> CreateAssistantTaskRunResult:
    task_id = uuid.uuid4().hex
    assistant_output = NATSAssistantUpdateSink(task_id=task_id, js=js)
    with logfire.span(
        "Customer Support Assistant",
        task_id=task_id,
    ):
        return await assistant_serve.create_assistant_task(
            task_id=task_id,
            assistant_awaitable=run_customer_support_agent(body),
            assistant_update_sink=assistant_output,
        )


@app.post("/tasks/{task_id}/message")
async def send_message_to_assistant_task(
    task_id: str,
    message: PostAssistantTaskMessage,
) -> Response:
    return await assistant_serve.send_message_to_task(task_id, message)


if __name__ == "__main__":
    import uvicorn

    uvicorn.run("examples.customer_support.app:app", host="0.0.0.0", port=8000)



from typing import Any

from tfy_assistant_framework.swarm import Agent, client


class Chat(Agent[Any]): ...


chat_agent = Chat(
    instructions="Chat with the user",
)


async def run_chat_assistant() -> None:
    async for agent_run in client.run(agent=chat_agent, messages=[], interactive=True):
        agent_run.debug_log()


async def main() -> None:
    await run_chat_assistant()


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())
